use futures::future;
use std::{
    collections::HashMap,
    error::Error,
    fmt::{self, Display},
    fs::read,
    io::Cursor,
    mem::swap,
    net::{IpAddr, SocketAddr},
    num::NonZeroU32,
    pin::Pin,
    str::FromStr,
    sync::{
        atomic::{AtomicBool, Ordering},
        Arc,
    },
    time::SystemTime,
};

use tokio::{
    sync::{oneshot, Mutex},
    time::{interval_at, Timeout},
};

use byteorder::{BigEndian, WriteBytesExt};
use env_logger::Env;
use itertools::Itertools;
use log::{debug, error, info, warn};
use nom::{bytes::complete::take, multi::many_m_n, number::complete::be_u8, IResult};
use nom_derive::{NomBE, Parse};
use proc_bitfield::*;
use rand::prelude::*;
use tokio::time;
use tokio::{io, join, task::JoinHandle};
use tokio::{net::UdpSocket, sync::RwLock};
use tokio::{sync::mpsc, time::Instant};
use tokio::{task, time::Interval};

use atomic::Atomic;
use bytemuck::{NoUninit, Pod};

const CONTROL_PORT: u16 = 3784;
const ECHO_PORT: u16 = 3785;

const ORDERING: Ordering = Ordering::Relaxed;

#[repr(u8)]
#[derive(ConvRaw, Debug, NomBE, PartialEq, Eq, Clone, Copy, NoUninit)]
pub enum BfdDiagnostic {
    None = 0,
    TimeExpired = 1,
    EchoFailed = 2,
    NeighborDown = 3,
    FwdPlaneReset = 4,
    PathDown = 5,
    ConcatPathDown = 6,
    AdminDown = 7,
    RevConcatPathDown = 8,
    Reserved,
}
impl Display for BfdDiagnostic {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str(match self {
            Self::None => "None",
            Self::TimeExpired => "TimeExpired",
            Self::EchoFailed => "EchoFailed",
            Self::NeighborDown => "NeighborDown",
            Self::FwdPlaneReset => "FwdPlaneReset",
            Self::PathDown => "PathDown",
            Self::ConcatPathDown => "ConcatPathDown",
            Self::AdminDown => "AdminDown",
            Self::RevConcatPathDown => "RevConcatPathDown",
            Self::Reserved => "Reserved",
        })
    }
}
#[repr(u8)]
#[derive(ConvRaw, Debug, NomBE, PartialEq, Eq, Default, Clone, Copy, NoUninit)]
pub enum BfdState {
    AdminDown = 0,
    #[default]
    Down = 1,
    Init = 2,
    Up = 3,
}
impl Display for BfdState {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            Self::AdminDown => "AdminDown",
            Self::Down => "Down",
            Self::Init => "Init",
            Self::Up => "Up",
        })
    }
}

#[repr(u8)]
#[derive(ConvRaw, Debug, NomBE, PartialEq, Eq, Clone, Copy, NoUninit)]
pub enum BfdAuthType {
    None = 0,
    SimplePassword = 1,
    KeyedMD5 = 2,
    MetKeyedMD5 = 3,
    KeyedSHA1 = 4,
    MetKeyedSHA1 = 5,
    Reserved,
}

#[derive(Debug)]
pub enum BfdError {
    // field, value
    InvalidFieldValue(&'static str, &'static str),
}
impl Display for BfdError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::InvalidFieldValue(field, value) => {
                write!(f, "invalid value `{}` for field `{}`", field, value)
            }
        }
    }
}
impl Error for BfdError {}

#[repr(transparent)]
#[derive(Debug, NomBE, PartialEq, Eq, Clone, Copy, NoUninit, Hash)]
pub struct BfdDiscriminator(u32);
impl Display for BfdDiscriminator {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

#[repr(transparent)]
#[derive(Debug, PartialEq, Eq, Clone, Copy, PartialOrd, Ord, NoUninit, NomBE)]
/// All intervals in BFD are specified in microseconds
pub struct BfdInterval(u32);
impl From<BfdInterval> for time::Duration {
    fn from(value: BfdInterval) -> Self {
        time::Duration::from_micros(value.0 as u64)
    }
}
impl From<u32> for BfdInterval {
    fn from(value: u32) -> Self {
        BfdInterval(value)
    }
}
bitfield! {
    #[derive(NomBE)]
    pub struct BfdFlags(pub u32): Debug {
        pub vers: u8 @ 29..=31,
        pub diag: u8 [try_get BfdDiagnostic] @ 24..=28,
        pub state: u8 [try_get BfdState] @ 22..=23,
        pub poll: bool @ 21,
        pub final_: bool @ 20,
        pub cpi: bool @ 19,
        pub auth_present: bool @ 18,
        pub demand: bool @ 17,
        pub multipoint: bool @ 16,
        pub detect_mult: u8 @ 8..=15,
        pub length: u8 @ 0..=7
    }
}

#[derive(Debug)]
pub struct BfdAuthSimplePassword(Vec<u8>);
impl<'a> Parse<&'a [u8]> for BfdAuthSimplePassword {
    fn parse(i: &'a [u8]) -> IResult<&'a [u8], Self, nom::error::Error<&'a [u8]>> {
        let (i, res) = many_m_n(1, 16, be_u8)(i)?;
        Ok((i, Self(res)))
    }
}

#[derive(Debug, NomBE)]
pub struct BfdAuthKeyedMD5 {
    key_id: u8,
    _reserved: u8,
    seq: u32,
    digest: [u8; 16],
}

#[derive(Debug, NomBE)]
pub struct BfdAuthKeyedSHA1 {
    key_id: u8,
    _reserved: u8,
    seq: u32,
    hash: [u8; 20],
}

#[derive(Debug, NomBE)]
#[nom(Selector = "BfdAuthType", Complete)]
pub enum BfdAuthData {
    #[nom(Selector = "BfdAuthType::SimplePassword")]
    SimplePassword(BfdAuthSimplePassword),
    #[nom(Selector = "BfdAuthType::KeyedMD5")]
    KeyedMD5(BfdAuthKeyedMD5),
    #[nom(Selector = "BfdAuthType::MetKeyedMD5")]
    MetKeyedMD5(BfdAuthKeyedMD5),
    #[nom(Selector = "BfdAuthType::KeyedSHA1")]
    KeyedSHA1(BfdAuthKeyedSHA1),
    #[nom(Selector = "BfdAuthType::MetKeyedSHA1")]
    MetKeyedSHA1(BfdAuthKeyedSHA1),
}

impl BfdAuthData {
    fn parse_be_with_length(
        i: &[u8],
        auth_type: BfdAuthType,
        auth_len: u8,
    ) -> IResult<&[u8], Self> {
        let (new_i, data) = take(auth_len)(i)?;
        let (_leftovers, retval) = BfdAuthData::parse_be(data, auth_type)?;
        Ok((new_i, retval))
    }
}

#[derive(Debug, NomBE)]
pub struct BfdAuth {
    auth_type: BfdAuthType,
    auth_len: u8,
    #[nom(Parse = "{ |i| BfdAuthData::parse_be_with_length(i, auth_type, auth_len) }")]
    auth_data: BfdAuthData,
}

#[derive(Debug, NomBE)]
pub struct BfdPacket {
    flags: BfdFlags,
    my_disc: BfdDiscriminator,
    your_disc: BfdDiscriminator,
    desired_min_tx: BfdInterval,
    required_min_rx: BfdInterval,
    required_min_echo_rx: BfdInterval,
    #[nom(Cond = "flags.auth_present()")]
    auth: Option<BfdAuth>,
}

impl BfdPacket {
    fn serialize(&self) -> Result<Box<[u8]>, std::io::Error> {
        // TODO: serialize auth
        let buf = [0u8; 24];
        let mut wtr = Cursor::new(buf);
        wtr.write_u32::<BigEndian>(self.flags.0)?;
        wtr.write_u32::<BigEndian>(self.my_disc.0)?;
        wtr.write_u32::<BigEndian>(self.your_disc.0)?;
        wtr.write_u32::<BigEndian>(self.desired_min_tx.0)?;
        wtr.write_u32::<BigEndian>(self.required_min_rx.0)?;
        wtr.write_u32::<BigEndian>(self.required_min_echo_rx.0)?;

        Ok(Box::new(wtr.into_inner()))
    }
}

/// Data structure to store the state of the Bfd machine. The impl does *not*
/// implement any Bfd logic, it is merely a thread-safe data structure
#[derive(Debug)]
struct BfdSessionState {
    control_sock: Arc<UdpSocket>,                  // immutable
    peer_addr: IpAddr,                             // immutable and Copy
    session_state: Atomic<BfdState>,               // mutable by any thread
    remote_session_state: Atomic<BfdState>,        // mutable by rx thread
    local_discr: BfdDiscriminator,                 // immutable and Copy
    remote_discr: Atomic<BfdDiscriminator>,        // mutable by rx thread
    local_diag: Atomic<BfdDiagnostic>,             // mutable by any thread
    desired_min_tx_interval: Atomic<BfdInterval>,  // mutable by any thread
    required_min_rx_interval: Atomic<BfdInterval>, // mutable by any thread
    remote_min_rx_interval: Atomic<BfdInterval>,   // mutable by rx thread
    demand_mode: Atomic<bool>,                     // mutable by rx thread
    remote_demand_mode: Atomic<bool>,              // mutable by rx thread
    detect_mult: Atomic<u8>,                       // mutable by control thread
    auth_type: Atomic<BfdAuthType>,                // mutable by control thread
    rcv_auth_seq: Atomic<u32>,                     // mutable by rx thread
    xmit_auth_seq: Atomic<u32>,                    // mutable by tx thread?
    auth_seq_known: Atomic<bool>,                  // mutable by rx thread?

    detection_time: Atomic<BfdInterval>, // mutable by rx thread
    poll_mode: Atomic<bool>,             // mutable by rx thread

    control_packets_rx: Atomic<u64>,
    control_packets_tx: Atomic<u64>,
    last_state_change: Mutex<Instant>,
    last_remote_diag: Atomic<BfdDiagnostic>,
}

// TODO: hide the internal data structure behind a trait
impl BfdSessionState {
    fn control_sock(&self) -> Arc<UdpSocket> {
        self.control_sock.clone()
    }
    fn peer_addr(&self) -> IpAddr {
        self.peer_addr
    }

    fn session_state(&self) -> BfdState {
        self.session_state.load(ORDERING)
    }
    // Note: This does not handle state transition
    fn set_session_state(&self, value: BfdState) {
        self.session_state.store(value, ORDERING)
    }
    fn remote_session_state(&self) -> BfdState {
        self.remote_session_state.load(ORDERING)
    }
    fn set_remote_session_state(&self, value: BfdState) {
        self.remote_session_state.store(value, ORDERING);
    }
    fn local_discr(&self) -> BfdDiscriminator {
        self.local_discr
    }
    fn remote_discr(&self) -> BfdDiscriminator {
        self.remote_discr.load(ORDERING)
    }
    fn set_remote_discr(&self, value: BfdDiscriminator) {
        self.remote_discr.store(value, ORDERING);
    }
    fn demand_mode(&self) -> bool {
        self.demand_mode.load(ORDERING)
    }
    fn set_demand_mode(&self, value: bool) {
        self.demand_mode.store(value, ORDERING)
    }
    fn remote_demand_mode(&self) -> bool {
        self.remote_demand_mode.load(ORDERING)
    }
    fn set_remote_demand_mode(&self, value: bool) {
        self.remote_demand_mode.store(value, ORDERING);
    }
    fn remote_min_rx_interval(&self) -> BfdInterval {
        self.remote_min_rx_interval.load(ORDERING)
    }
    fn set_remote_min_rx_interval(&self, value: BfdInterval) {
        self.remote_min_rx_interval.store(value, ORDERING);
    }
    fn local_diag(&self) -> BfdDiagnostic {
        self.local_diag.load(ORDERING)
    }
    fn set_local_diag(&self, value: BfdDiagnostic) {
        self.local_diag.store(value, ORDERING);
    }
    fn auth_type(&self) -> BfdAuthType {
        self.auth_type.load(ORDERING)
    }
    fn detect_mult(&self) -> u8 {
        self.detect_mult.load(ORDERING)
    }
    fn set_detect_mult(&self, value: u8) {
        self.detect_mult.store(value, ORDERING)
    }
    fn poll_mode(&self) -> bool {
        self.poll_mode.load(ORDERING)
    }
    fn set_poll_mode(&self, value: bool) {
        self.poll_mode.store(value, ORDERING)
    }
    fn desired_min_tx_interval(&self) -> BfdInterval {
        self.desired_min_tx_interval.load(ORDERING)
    }
    fn set_desired_min_tx_interval(&self, value: BfdInterval) {
        self.desired_min_tx_interval.store(value, ORDERING)
    }
    fn required_min_rx_interval(&self) -> BfdInterval {
        self.required_min_rx_interval.load(ORDERING)
    }
    fn set_required_min_rx_interval(&self, value: BfdInterval) {
        self.required_min_rx_interval.store(value, ORDERING)
    }
    fn detection_time(&self) -> BfdInterval {
        self.detection_time.load(ORDERING)
    }
    fn set_detection_time(&self, value: BfdInterval) {
        self.detection_time.store(value, ORDERING)
    }
    fn control_packets_rx(&self) -> u64 {
        self.control_packets_rx.load(ORDERING)
    }
    fn control_packets_tx(&self) -> u64 {
        self.control_packets_tx.load(ORDERING)
    }
    fn last_state_change(&self) -> Instant {
        *self.last_state_change.try_lock().unwrap()
    }
    fn last_remote_diag(&self) -> BfdDiagnostic {
        self.last_remote_diag.load(ORDERING)
    }
    fn set_last_remote_diag(&self, value: BfdDiagnostic) {
        self.last_remote_diag.store(value, ORDERING)
    }
}

enum WatchdogReset {
    ReceivedPacket,
    Terminate,
}

struct BfdSession {
    state: BfdSessionState,
    periodic: BfdPeriodicSenderHandle,
    rx: mpsc::Receiver<SessionControlCommand>,
    tx: mpsc::Sender<SessionControlCommand>,
    rx_watchdog: Option<mpsc::Sender<WatchdogReset>>,
    cur_interval: BfdInterval,
}

#[derive(Debug)]
struct BfdSessionStats {
    local_ip: IpAddr,
    remote_ip: IpAddr,
    local_discr: BfdDiscriminator,
    remote_discr: BfdDiscriminator,
    state: BfdState,
    last_diag: BfdDiagnostic,
    control_packets_rx: u64,
    control_packets_tx: u64,
    last_change: Instant,
}
impl Display for BfdSessionStats {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        writeln!(
            f,
            "{:<20} {:<9} {:>6} {:>6}",
            self.remote_ip, self.state, self.control_packets_rx, self.control_packets_tx
        )?;
        write!(f, "  LD:{} RD:{} Diag:{} Last: {}s", self.local_discr, self.remote_discr, self.last_diag, Instant::now().duration_since(self.last_change).as_secs())
    }
}
impl From<&BfdSessionState> for BfdSessionStats {
    fn from(state: &BfdSessionState) -> Self {
        Self {
            local_ip: state.control_sock().local_addr().unwrap().ip(),
            remote_ip: state.peer_addr(),
            local_discr: state.local_discr(),
            remote_discr: state.remote_discr(),
            state: state.session_state(),
            last_diag: state.last_remote_diag(),
            control_packets_rx: state.control_packets_rx(),
            control_packets_tx: state.control_packets_tx(),
            last_change: state.last_state_change(),
        }
    }
}
enum PeriodicControlCommand {
    Stop,
    Start,
    Quit,
    SetMinInterval(BfdInterval),
    SendPollResponse,
}

enum SessionControlCommand {
    RxPacket(BfdPacket),
    TxControlPacket {
        poll_response: bool,
    },
    GetSessionStats {
        respond_to: oneshot::Sender<BfdSessionStats>,
    },
    ReceiveTimeout,
    Quit,
}

#[derive(Clone)]
struct BfdSessionHandle {
    tx: mpsc::Sender<SessionControlCommand>,
}

impl BfdSessionHandle {
    async fn new(local_addr: IpAddr, remote_addr: IpAddr, local_discr: BfdDiscriminator) -> Self {
        let (tx, rx) = mpsc::channel(32);
        let new_self = Self { tx };
        let periodic = BfdPeriodicSenderHandle::new(new_self.clone(), BfdInterval(1_000_000));

        let mut session = BfdSession::new(
            rx,
            new_self.tx.clone(),
            periodic,
            local_addr,
            remote_addr,
            local_discr,
        )
        .await;
        tokio::spawn(async move { session.run().await });

        new_self
    }
    async fn transmit_control_packet(
        &self,
        poll_response: bool,
    ) -> Result<(), mpsc::error::SendError<SessionControlCommand>> {
        self.tx
            .send(SessionControlCommand::TxControlPacket { poll_response })
            .await
    }
    async fn get_stats(
        &self,
    ) -> Result<BfdSessionStats, mpsc::error::SendError<SessionControlCommand>> {
        let (send, recv) = oneshot::channel();
        self.tx
            .send(SessionControlCommand::GetSessionStats { respond_to: send })
            .await?;
        Ok(recv.await.expect("Unable to receive stats from session"))
    }
}

#[derive(Clone, Debug)]
struct BfdPeriodicSenderHandle {
    tx: mpsc::Sender<PeriodicControlCommand>,
}

impl BfdPeriodicSenderHandle {
    fn new(bfd: BfdSessionHandle, interval: BfdInterval) -> Self {
        let (tx, rx) = mpsc::channel(32);
        let mut periodic_sender = BfdPeriodicSender::new(rx, bfd, interval);
        tokio::spawn(async move { periodic_sender.run().await });

        Self { tx }
    }
}

struct BfdPeriodicSender {
    rx: mpsc::Receiver<PeriodicControlCommand>,
    bfd: BfdSessionHandle,
    cur_interval: BfdInterval,
}

impl BfdPeriodicSender {
    fn new(
        rx: mpsc::Receiver<PeriodicControlCommand>,
        bfd: BfdSessionHandle,
        interval: BfdInterval,
    ) -> Self {
        Self {
            rx,
            bfd,
            cur_interval: interval,
        }
    }
    async fn run(&mut self) {
        let base_interval = time::Duration::from_micros(self.cur_interval.0 as u64 * 3 / 4);
        let mut clock = time::interval(base_interval);
        let mut running = true;

        loop {
            tokio::select! {
             _ = clock.tick(), if running == true => {
                        let jitter = time::Duration::from_micros(
                            rand::thread_rng()
                                .gen_range(0..clock.period().as_micros() / 3)
                                .try_into()
                                .unwrap(),
                        );
                        time::sleep(jitter).await;
                        // debug!("tick! sent periodic control packet after {}ms", (base_interval + jitter).as_millis());
                        self.bfd.transmit_control_packet(false).await.unwrap();
                    },
                    Some(cmd) = self.rx.recv() => {
                        match cmd {
                            PeriodicControlCommand::Quit => { debug!("Closing periodic packet sender"); return},
                            PeriodicControlCommand::Stop => {
                                debug!("Stopping periodic packets");
                                running= false}
                            ,
                            PeriodicControlCommand::Start => { debug!("Starting periodic packets"); running = true },
                            PeriodicControlCommand::SetMinInterval(i) => {
                                let base_interval = time::Duration::from_micros(i.0 as u64 * 3 / 4);
                                debug!("Updating base interval to {}ms (jittering {} to {}ms)", base_interval.as_millis(), base_interval.as_millis(), base_interval.as_millis() + base_interval.as_millis() / 3);
                                clock = time::interval_at(
                                    time::Instant::now() + base_interval.into(),
                                    base_interval.into(),
                                );
                            },
                            PeriodicControlCommand::SendPollResponse => {
                                debug!("Sending poll response");
                                self.bfd.transmit_control_packet(true).await.unwrap();
                            }
                        }
                    }
            };
        }
    }
}

impl BfdSession {
    async fn new(
        rx: mpsc::Receiver<SessionControlCommand>,
        tx: mpsc::Sender<SessionControlCommand>,
        periodic: BfdPeriodicSenderHandle,
        local_addr: IpAddr,
        remote_addr: IpAddr,
        local_discr: BfdDiscriminator,
    ) -> Self {
        let mut rng = rand::thread_rng();

        //TODO: select a random unused port instead of pure random
        let source_port: u16 = rng.gen_range(49152..=65535);
        let control_sock = UdpSocket::bind(SocketAddr::new(local_addr, source_port))
            .await
            .unwrap();
        control_sock.set_ttl(255).unwrap();
        // control_sock
        //     .connect(SocketAddr::new(remote_addr, CONTROL_PORT))
        //     .await?;

        // Incoming packets will come over the channel from the mux, since they don't send to the reciprocal port
        Self {
            rx,
            tx,
            periodic,
            rx_watchdog: None,
            cur_interval: BfdInterval(1_000_000),
            state: BfdSessionState {
                control_sock: Arc::new(control_sock),
                peer_addr: remote_addr,
                session_state: Atomic::new(BfdState::default()),
                remote_session_state: Atomic::new(BfdState::default()),
                local_discr,
                remote_discr: Atomic::new(BfdDiscriminator(0)),
                local_diag: Atomic::new(BfdDiagnostic::None),
                desired_min_tx_interval: Atomic::new(BfdInterval(1_000_000)),
                required_min_rx_interval: Atomic::new(BfdInterval(300_000)),
                remote_min_rx_interval: Atomic::new(BfdInterval(1)),
                demand_mode: Atomic::new(false),
                remote_demand_mode: Atomic::new(false),
                detect_mult: Atomic::new(3),
                auth_type: Atomic::new(BfdAuthType::None),
                rcv_auth_seq: Atomic::new(0),
                xmit_auth_seq: Atomic::new(rng.gen()),
                auth_seq_known: Atomic::new(false),
                detection_time: Atomic::new(BfdInterval(0)),
                poll_mode: Atomic::new(false),

                control_packets_rx: Atomic::new(0),
                control_packets_tx: Atomic::new(0),
                last_state_change: Mutex::new(Instant::now()),
                last_remote_diag: Atomic::new(BfdDiagnostic::None),
            },
        }
    }

    async fn transition(&mut self, new_state: BfdState) {
        if self.state.session_state() == new_state {
            return;
        };
        info!(
            "Peer {} state change {} -> {}",
            self.state.peer_addr(),
            self.state.session_state(),
            new_state
        );
        *self.state.last_state_change.lock().await = Instant::now();
        match (self.state.session_state(), new_state) {
            (BfdState::Up, BfdState::AdminDown | BfdState::Down | BfdState::Init) => {
                if self.rx_watchdog.is_some() {
                    self.rx_watchdog
                        .as_mut()
                        .unwrap()
                        .send(WatchdogReset::Terminate)
                        .await
                        .unwrap();
                }
            }
            (BfdState::AdminDown | BfdState::Down | BfdState::Init, BfdState::Up) => {
                self.start_watchdog().await;
            }
            _ => {}
        }
        self.state.set_session_state(new_state);
    }

    async fn run(&mut self) {
        while let Some(cmd) = self.rx.recv().await {
            match cmd {
                SessionControlCommand::Quit => return,
                SessionControlCommand::RxPacket(packet) => {
                    self.receive_control_packet(&packet).await;
                }
                SessionControlCommand::GetSessionStats { respond_to } => {
                    respond_to.send(BfdSessionStats::from(&self.state)).unwrap()
                }
                SessionControlCommand::TxControlPacket { poll_response } => {
                    self.transmit_control_packet(poll_response).await
                }
                SessionControlCommand::ReceiveTimeout => {
                    self.state.set_local_diag(BfdDiagnostic::TimeExpired);
                    self.transition(BfdState::Down).await;
                }
            }
        }
        println!("Session command channel closed");
    }

    async fn transmit_control_packet(&self, poll_response: bool) {
        let packet = BfdPacket {
            flags: BfdFlags(0)
                .with_vers(1)
                .with_diag(self.state.local_diag().into())
                .with_state(self.state.session_state().into())
                .with_poll(if poll_response {
                    false
                } else {
                    self.state.poll_mode()
                })
                .with_final_(poll_response)
                .with_cpi(true)
                .with_demand(
                    self.state.session_state() == BfdState::Up
                        && self.state.remote_session_state() == BfdState::Up,
                )
                .with_detect_mult(self.state.detect_mult())
                .with_length(24),
            my_disc: self.state.local_discr(),
            your_disc: self.state.remote_discr(),
            desired_min_tx: self.state.desired_min_tx_interval(),
            required_min_rx: self.state.required_min_rx_interval(),
            required_min_echo_rx: BfdInterval(0),
            auth: None,
        };
        let socket = self.state.control_sock.clone();
        let dest = self.state.peer_addr;
        self.state.control_packets_tx.fetch_add(1, ORDERING);
        debug!("tx packet: {:?}", packet);
        socket
            .send_to(
                packet.serialize().unwrap().as_ref(),
                SocketAddr::new(dest, CONTROL_PORT),
            )
            .await
            .unwrap();
    }

    async fn start_watchdog(&mut self) {
        let (tx, mut rx) = mpsc::channel(1);
        let cmd_tx = self.tx.clone();
        let duration = self.state.detection_time().into();
        let peer_ip = self.state.peer_addr();

        tokio::spawn(async move {
            let mut running = true;
            while running {
                match tokio::time::timeout(duration, async { rx.recv().await }).await {
                    Ok(Some(WatchdogReset::ReceivedPacket)) => {
                        debug!("Packet received, watchdog ok")
                    }
                    Ok(Some(WatchdogReset::Terminate)) | Ok(None) => {
                        debug!("Watchdog terminating");
                        running = false
                    }
                    Err(_) => {
                        info!(
                            "Peer {} timed out after {}ms",
                            peer_ip,
                            duration.as_millis()
                        );
                        cmd_tx
                            .send(SessionControlCommand::ReceiveTimeout)
                            .await
                            .unwrap()
                    }
                }
            }
        });
        let mut temporary = Some(tx);
        swap(&mut temporary, &mut self.rx_watchdog);
        if temporary.is_some() {
            debug!("Updating old watchdog with new detection time");
            temporary
                .as_mut()
                .unwrap()
                .send(WatchdogReset::Terminate)
                .await
                .unwrap_or(())
        }
    }

    // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.6
    async fn receive_control_packet(&mut self, p: &BfdPacket) {
        // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.4
        //
        // If Demand mode is not active, and a period of time equal to the
        // Detection Time passes without receiving a BFD Control packet from the
        // remote system, and bfd.SessionState is Init or Up, the session has
        // gone down -- the local system MUST set bfd.SessionState to Down and
        // bfd.LocalDiag to 1 (Control Detection Time Expired).
        if self.rx_watchdog.is_some() {
            let _ = self
                .rx_watchdog
                .as_ref()
                .unwrap()
                .send(WatchdogReset::ReceivedPacket)
                .await;
        }
        self.state.control_packets_rx.fetch_add(1, ORDERING);
        let received_state = match p.flags.state() {
            Err(_) => {
                warn!("Invalid state, discarding");
                return;
            }
            Ok(v) => v,
        };
        // If the version number is not correct (1), the packet MUST be discarded.
        if p.flags.vers() != 1 {
            warn!("Invalid version {}, discarding", p.flags.vers());
            return;
        }
        //   If the Length field is less than the minimum correct value (24 if the A bit is clear, or 26 if the A bit is
        //   set), the packet MUST be discarded.
        if p.flags.length() < 24 || (p.flags.length() < 26 && p.flags.auth_present()) {
            warn!("Invalid packet length {}, discarding", p.flags.length());
            return;
        }
        // TODO: If the Length field is greater than the payload of the encapsulating protocol, the packet MUST be
        //   discarded.

        // If the Detect Mult field is zero, the packet MUST be discarded.
        if p.flags.detect_mult() == 0 {
            warn!("Invalid detect mult {}, discarding", p.flags.detect_mult());
            return;
        }

        //If the Multipoint (M) bit is nonzero, the packet MUST be discarded.
        if p.flags.multipoint() {
            warn!("Invalid multipoint enabled, discarding");
            return;
        }

        // If the My Discriminator field is zero, the packet MUST be discarded.
        if p.my_disc == BfdDiscriminator(0) {
            warn!("Invalid my discriminator {:?}, discarding", p.my_disc);
            return;
        }

        // If the Your Discriminator field is nonzero, it MUST be used to select the session with which this BFD packet
        //   is associated.  If no session is found, the packet MUST be discarded.
        if p.your_disc != BfdDiscriminator(0) && p.your_disc != self.state.local_discr() {
            warn!(
                "Received unexpected discriminator {:?}, discarding",
                p.your_disc
            );
            return;
        }

        // If the Your Discriminator field is zero and the State field is not Down or AdminDown, the packet MUST be
        //   discarded.
        if p.your_disc == BfdDiscriminator(0)
            && (received_state != BfdState::Down && received_state != BfdState::AdminDown)
        {
            warn!(
                "Got packet with zero discriminator and invalid state {:?}, discarding",
                received_state
            );
            return;
        }

        // If the A bit is set and no authentication is in use (bfd.AuthType is zero), the packet MUST be discarded.
        if p.flags.auth_present() && self.state.auth_type() == BfdAuthType::None {
            warn!("Got packet with auth enabled when we disagree, discarding");
            return;
        }

        // If the A bit is clear and authentication is in use (bfd.AuthType is nonzero), the packet MUST be discarded.
        if !p.flags.auth_present() && self.state.auth_type() != BfdAuthType::None {
            warn!("Got packet without auth when we expect it, discarding");
            return;
        }

        // If the A bit is set, the packet MUST be authenticated under the rules of section 6.7, based on the
        //   authentication type in use (bfd.AuthType).  This may cause the packet to be discarded.
        if p.flags.auth_present() {
            unimplemented!("Authentication is not implemented");
        }

        // Set bfd.RemoteDiscr to the value of My Discriminator.
        self.state.set_remote_discr(p.my_disc);
        // Set bfd.RemoteState to the value of the State (Sta) field.
        self.state.set_remote_session_state(received_state);
        // Set bfd.RemoteDemandMode to the value of the Demand (D) bit.
        self.state.set_remote_demand_mode(p.flags.demand());
        // Set bfd.RemoteMinRxInterval to the value of Required Min RX Interval.
        self.state.set_remote_min_rx_interval(p.required_min_rx);
        // Not in spec: store the received diagnostic
        self.state.set_last_remote_diag(p.flags.diag().unwrap());

        // If the Required Min Echo RX Interval field is zero, the transmission of Echo packets, if any, MUST cease.
        if p.required_min_echo_rx == BfdInterval(0) {
            // TODO: implement echo thread
        }
        // If a Poll Sequence is being transmitted by the local system and the Final (F) bit in the received packet is
        //   set, the Poll Sequence MUST be terminated.
        //
        // TODO: poll stuff

        // Update the transmit interval as described in section 6.8.2.
        self.update_transmit_interval().await;

        // Update the Detection Time as described in section 6.8.4.
        self.update_detection_time(&p).await;

        // If bfd.SessionState is AdminDown
        //   Discard the packet
        if self.state.session_state() == BfdState::AdminDown {
            return;
        }

        // If received state is AdminDown
        if received_state == BfdState::AdminDown {
            // If bfd.SessionState is not Down
            if self.state.session_state() != BfdState::Down {
                // Set bfd.LocalDiag to 3 (Neighbor signaled session down)
                self.state.set_local_diag(BfdDiagnostic::NeighborDown);
                // Set bfd.SessionState to Down
                self.transition(BfdState::Down).await;
            }
        } else {
            // If bfd.SessionState ...
            match self.state.session_state() {
                BfdState::Down => {
                    // If received State is Down
                    if received_state == BfdState::Down {
                        // Set bfd.SessionState to Init
                        self.transition(BfdState::Init).await;
                    // Else if received State is Init
                    } else if received_state == BfdState::Init {
                        // Set bfd.SessionState to Up
                        self.transition(BfdState::Up).await;
                    }
                }
                BfdState::Init => {
                    // If received State is Init or Up
                    if received_state == BfdState::Init || received_state == BfdState::Up {
                        // Set bfd.SessionState to Up
                        self.transition(BfdState::Up).await;
                    }
                }
                BfdState::Up => {
                    // If received State is Down
                    if received_state == BfdState::Down {
                        // Set bfd.LocalDiag to 3 (Neighbor signaled session down)
                        self.state.set_local_diag(BfdDiagnostic::NeighborDown);
                        // Set bfd.SessionState to Down
                        self.transition(BfdState::Down).await;
                    }
                }
                BfdState::AdminDown => unreachable!("unexpected AdminDown"), // AdminDown is discarded earlier
            }

            // Check to see if Demand mode should become active or not (see section 6.6).
            // - we don't use demand mode

            // If bfd.RemoteDemandMode is 1, bfd.SessionState is Up, and bfd.RemoteSessionState is Up, Demand mode is
            // active on the remote system and
            if self.state.remote_demand_mode()
                && self.state.session_state() == BfdState::Up
                && self.state.remote_session_state() == BfdState::Up
            {
                // the local system MUST cease the periodic transmission of BFD Control packets (see section 6.8.7).
                self.periodic
                    .tx
                    .send(PeriodicControlCommand::Stop)
                    .await
                    .unwrap();
            }

            // If the Poll (P) bit is set, send a BFD Control packet to the remote system with the Poll (P) bit clear,
            // and the Final (F) bit set (see section 6.8.7).
            if p.flags.poll() {}
            // If the packet was not discarded, it has been received for purposes of the Detection Time expiration rules
            // in section 6.8.4.
        }
    }

    // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.2
    async fn update_transmit_interval(&mut self) {
        let new_base_interval = std::cmp::max(
            self.state.desired_min_tx_interval(),
            self.state.remote_min_rx_interval(),
        );
        if new_base_interval != self.cur_interval {
            self.periodic
                .tx
                .send(PeriodicControlCommand::SetMinInterval(new_base_interval))
                .await
                .unwrap()
        }
        self.cur_interval = new_base_interval;
    }

    // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.4
    async fn update_detection_time(&mut self, p: &BfdPacket) {
        let old = self.state.detection_time();
        self.state.set_detection_time(if !self.state.demand_mode() {
            (p.flags.detect_mult() as u32
                * std::cmp::max(self.state.required_min_rx_interval(), p.desired_min_tx).0)
                .into()
        } else {
            (self.state.detect_mult() as u32
                * std::cmp::max(
                    self.state.desired_min_tx_interval(),
                    self.state.remote_min_rx_interval(),
                )
                .0)
                .into()
        });
        if old != self.state.detection_time() {
            self.start_watchdog().await
        }
    }
}

#[tokio::main]
async fn main() -> io::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();

    info!("rust-bfd starting up");

    let local = SocketAddr::from_str("192.168.122.1:3784").unwrap();
    let peers = vec![SocketAddr::from_str("192.168.122.132:3784").unwrap()];
    let mut rng = rand::thread_rng();
    let mut sessions_by_ip = HashMap::new();
    let mut sessions_by_discr = HashMap::new();

    for peer in peers {
        let mut local_discr = BfdDiscriminator(0);
        while local_discr.0 == 0 || sessions_by_discr.get(&local_discr).is_some() {
            local_discr = BfdDiscriminator(rng.gen());
        }
        let session = BfdSessionHandle::new(local.ip(), peer.ip(), local_discr).await;

        sessions_by_ip.insert((local.ip(), peer.ip()), local_discr);
        sessions_by_discr.insert(local_discr, session);
    }
    debug!(
        "configured sessions: {}",
        future::join_all(
            sessions_by_discr
                .values()
                .map(|session| async { session.get_stats().await.unwrap().to_string() })
        )
        .await
        .join("\n\t")
    );

    let control_sock = Arc::new(UdpSocket::bind(local).await?);
    // If BFD authentication is not in use on a session, all BFD Control packets for the session MUST be sent with a
    // Time to Live (TTL) or Hop Limit value of 255.
    control_sock.set_ttl(255).unwrap();
    let echo_socket = Arc::new(UdpSocket::bind(SocketAddr::new(local.ip(), ECHO_PORT)).await?);

    let rx_thread = task::spawn(async move {
        let mut buf = [0; 1024];
        loop {
            // TODO: All received BFD Control packets that are demultiplexed to the session MUST be discarded if the
            // received TTL or Hop Limit is not equal to 255. Need raw sockets for this.
            let (len, addr) = control_sock.recv_from(&mut buf).await.unwrap(); // TODO: fallibility?
            if let Ok((_leftover, packet)) = BfdPacket::parse(buf.as_slice()) {
                debug!("rx packet: {:?}", packet);
                let our_disc = if packet.your_disc == BfdDiscriminator(0) {
                    if let Some(discr) =
                        sessions_by_ip.get(&(control_sock.local_addr().unwrap().ip(), addr.ip()))
                    {
                        debug!(
                            "Found session for unknown discriminator from {}: {}",
                            addr.ip(),
                            discr
                        );
                        discr
                    } else {
                        warn!("Unable to match packet to session with {}", addr.ip());
                        continue;
                    }
                } else {
                    &packet.your_disc
                };
                if let Some(session) = sessions_by_discr.get(our_disc) {
                    session
                        .tx
                        .send(SessionControlCommand::RxPacket(packet))
                        .await
                        .unwrap();
                } else {
                    warn!("Unable to match packet to session with {}", our_disc);
                    continue;
                }
            } else {
                warn!("Failed to parse packet");
            }

            debug!(
                "configured sessions: {}",
                future::join_all(
                    sessions_by_discr
                        .values()
                        .map(|session| async { session.get_stats().await.unwrap().to_string() })
                )
                .await
                .join("\n\t")
            );
        }
    });

    tokio::join!(rx_thread)
        .0
        .expect("Unable to join on the receive thread");

    Ok(())
}
